@@ -15,6 +15,9 @@
 
 #include "common.h"
 #include "scrext.h"
+#if defined(MYFCGI)
+#include <fcgi_stdio.h>
+#endif
 
 #define SALTNAME       "[salt]"          // dummy user name of the salt
 #define RIDDLENAME     "[riddle]"        // dummy user name of the riddle
@@ -29,18 +32,20 @@ typedef struct {                         // type of structure for a record
 
 
 /* global variables */
+TCMPOOL *g_mpool;                        // global memory pool
+TCTMPL *g_tmpl;                          // template serializer
+void *g_scrextproc;                      // processor of the script extension
+TCMAP *g_users;                          // user list
+long g_eventcount;                       // event counter
 const char *g_scriptname;                // script name
 const char *g_scriptprefix;              // script prefix
 const char *g_scriptpath;                // script path
 const char *g_docroot;                   // document root
-TCTMPL *g_tmpl;                          // template serializer
 const char *g_database;                  // path of the database file
 const char *g_password;                  // path of the password file
-TCMAP *g_users;                          // user list
 const char *g_upload;                    // path of the upload directory
 const char *g_uploadpub;                 // public path of the upload directory
 const char *g_scrext;                    // path of the script extension file
-void *g_scrextproc;                      // processor of the script extension
 int64_t g_recvmax;                       // maximum size of received data
 const char *g_title;                     // site title
 int g_searchnum;                         // number of articles in a search page
@@ -56,11 +61,11 @@ const char *g_frontpage;                 // name of the front page
 
 /* function prototypes */
 int main(int argc, char **argv);
+static int realmain(int argc, char **argv);
 static void showerror(int code, const char *msg);
 static void showcache(void);
 static void readpasswd(void);
 static bool writepasswd(void);
-static void eventloop(void);
 static void dosession(TCMPOOL *mpool);
 static void setdberrmsg(TCLIST *emsgs, TCTDB *tdb, const char *msg);
 static void setarthtml(TCMPOOL *mpool, TCMAP *cols, int64_t id, int bhl, bool tiny);
@@ -79,112 +84,121 @@ static bool doupdatecmd(TCMPOOL *mpool, const char *mode, const char *baseurl, c
 
 /* main routine */
 int main(int argc, char **argv){
+#if defined(MYFCGI)
+  g_mpool = tcmpoolglobal();
+  int rv = 0;
+  while(FCGI_Accept() >= 0){
+    g_eventcount++;
+    if(realmain(argc, argv) != 0) rv = 1;
+  }
+  return rv;
+#else
+  g_mpool = tcmpoolglobal();
+  return realmain(argc, argv);
+#endif
+}
+
+
+/* real main routine */
+static int realmain(int argc, char **argv){
+  TCMPOOL *mpool = tcmpoolnew();
   g_scriptname = getenv("SCRIPT_NAME");
   if(!g_scriptname) g_scriptname = argv[0];
-  char *prefix = tcregexreplace(g_scriptname, "\\.[a-zA-Z0-9]*$", "");
+  char *prefix = tcmpoolpushptr(mpool, tcregexreplace(g_scriptname, "\\.[a-zA-Z0-9]*$", ""));
   g_scriptprefix = prefix;
   g_scriptpath = getenv("SCRIPT_FILENAME");
   if(!g_scriptpath) g_scriptpath = argv[0];
   g_docroot = getenv("DOCUMENT_ROOT");
-  char *docrootbuf = NULL;
   if(!g_docroot){
     g_docroot = "/";
     if(*g_scriptpath == '/'){
       int diff = strlen(g_scriptpath) - strlen(g_scriptname);
-      if(diff > 0 && !strcmp(g_scriptpath + diff, g_scriptname)){
-        docrootbuf = tcmemdup(g_scriptpath, diff);
-        g_docroot = docrootbuf;
-      }
+      if(diff > 0 && !strcmp(g_scriptpath + diff, g_scriptname))
+        g_docroot = tcmpoolpushptr(mpool, tcmemdup(g_scriptpath, diff));
     }
   }
-  char *barepath = tcregexreplace(g_scriptpath, "\\.[a-zA-Z0-9]*$", "");
-  char *tmplpath = tcsprintf("%s.tmpl", barepath);
+  char *barepath = tcmpoolpushptr(mpool, tcregexreplace(g_scriptpath, "\\.[a-zA-Z0-9]*$", ""));
+  char *tmplpath = tcmpoolpushptr(mpool, tcsprintf("%s.tmpl", barepath));
   const char *rp = getenv("REQUEST_URI");
-  char *scriptnamebuf = NULL;
   if(rp && *rp == '/'){
-    scriptnamebuf = strdup(rp);
-    char *pv = strchr(scriptnamebuf, '#');
+    char *buf = tcmpoolpushptr(mpool, strdup(rp));
+    char *pv = strchr(buf, '#');
     if(pv) *pv = '\0';
-    pv = strchr(scriptnamebuf, '?');
+    pv = strchr(buf, '?');
     if(pv) *pv = '\0';
-    if(*scriptnamebuf != '\0') g_scriptname = scriptnamebuf;
+    if(*buf != '\0') g_scriptname = buf;
   }
-  g_tmpl = tctmplnew();
-  if(tctmplload2(g_tmpl, tmplpath)){
-    g_database = tctmplconf(g_tmpl, "database");
-    if(!g_database) g_database = "promenade.tct";
-    g_password = tctmplconf(g_tmpl, "password");
-    g_users = NULL;
-    if(g_password){
-      g_users = tcmapnew2(TINYBNUM);
-      readpasswd();
-    }
-    g_upload = tctmplconf(g_tmpl, "upload");
-    g_uploadpub = NULL;
-    char *uploadpubbuf = NULL;
-    if(g_upload && *g_upload != '\0'){
-      if(!strchr(g_upload, '/')){
-        g_uploadpub = g_upload;
-      } else {
-        uploadpubbuf = tcrealpath(g_upload);
-        if(uploadpubbuf){
-          if(tcstrfwm(uploadpubbuf, g_docroot)){
-            int plen = strlen(g_docroot);
-            if(uploadpubbuf[plen] == '/') g_uploadpub = uploadpubbuf + plen;
+  if(g_tmpl){
+    dosession(mpool);
+  } else {
+    g_tmpl = tcmpoolpush(g_mpool, tctmplnew(), (void (*)(void *))tctmpldel);
+    if(tctmplload2(g_tmpl, tmplpath)){
+      g_database = tctmplconf(g_tmpl, "database");
+      if(!g_database) g_database = "promenade.tct";
+      g_password = tctmplconf(g_tmpl, "password");
+      g_users = NULL;
+      if(g_password){
+        g_users = tcmpoolpushmap(g_mpool, tcmapnew2(TINYBNUM));
+        readpasswd();
+      }
+      g_upload = tctmplconf(g_tmpl, "upload");
+      g_uploadpub = NULL;
+      if(g_upload && *g_upload != '\0'){
+        if(!strchr(g_upload, '/')){
+          g_uploadpub = g_upload;
+        } else {
+          char *rpath = tcmpoolpushptr(g_mpool, tcrealpath(g_upload));
+          if(rpath){
+            if(tcstrfwm(rpath, g_docroot)){
+              int plen = strlen(g_docroot);
+              if(rpath[plen] == '/') g_uploadpub = rpath + plen;
+            }
           }
         }
       }
+      g_scrext = tctmplconf(g_tmpl, "scrext");
+      g_scrextproc = tcmpoolpush(g_mpool, g_scrext && *g_scrext != '\0' ? scrextnew() : NULL,
+                                 scrextdel);
+      const char *rp = tctmplconf(g_tmpl, "recvmax");
+      g_recvmax = rp ? tcatoix(rp) : INT_MAX;
+      g_title = tctmplconf(g_tmpl, "title");
+      if(!g_title) g_title = "Tokyo Promenade";
+      rp = tctmplconf(g_tmpl, "searchnum");
+      g_searchnum = tclmax(rp ? tcatoi(rp) : 10, 1);
+      rp = tctmplconf(g_tmpl, "listnum");
+      g_listnum = tclmax(rp ? tcatoi(rp) : 10, 1);
+      rp = tctmplconf(g_tmpl, "feedlistnum");
+      g_feedlistnum = tclmax(rp ? tcatoi(rp) : 10, 1);
+      rp = tctmplconf(g_tmpl, "filenum");
+      g_filenum = tclmax(rp ? tcatoi(rp) : 10, 1);
+      rp = tctmplconf(g_tmpl, "sidebarnum");
+      g_sidebarnum = tclmax(rp ? tcatoi(rp) : 0, 0);
+      g_commentmode = tctmplconf(g_tmpl, "commentmode");
+      if(!g_commentmode) g_commentmode = "";
+      g_updatecmd = tctmplconf(g_tmpl, "updatecmd");
+      if(!g_updatecmd) g_updatecmd = "";
+      rp = tctmplconf(g_tmpl, "sessionlife");
+      g_sessionlife = tclmax(rp ? tcatoi(rp) : 0, 0);
+      g_frontpage = tctmplconf(g_tmpl, "frontpage");
+      if(!g_frontpage) g_frontpage = "";
+      TCMAP *conf = g_tmpl->conf;
+      tcmapiterinit(conf);
+      while((rp = tcmapiternext2(conf)) != NULL){
+        const char *pv = tcmapiterval2(rp);
+        char *name = tcmpoolpushptr(g_mpool, tcsprintf("TP_%s", rp));
+        tcstrtoupper(name);
+        setenv(name, pv, 1);
+      }
+      if(g_scrextproc){
+        scrextsetmapvar(g_scrextproc, "_conf", conf);
+        scrextload(g_scrextproc, g_scrext);
+      }
+      dosession(mpool);
+    } else {
+      showerror(500, "The template file is missing.");
     }
-    g_scrext = tctmplconf(g_tmpl, "scrext");
-    g_scrextproc = g_scrext && *g_scrext != '\0' ? scrextnew() : NULL;
-    const char *rp = tctmplconf(g_tmpl, "recvmax");
-    g_recvmax = rp ? tcatoix(rp) : INT_MAX;
-    g_title = tctmplconf(g_tmpl, "title");
-    if(!g_title) g_title = "Tokyo Promenade";
-    rp = tctmplconf(g_tmpl, "searchnum");
-    g_searchnum = tclmax(rp ? tcatoi(rp) : 10, 1);
-    rp = tctmplconf(g_tmpl, "listnum");
-    g_listnum = tclmax(rp ? tcatoi(rp) : 10, 1);
-    rp = tctmplconf(g_tmpl, "feedlistnum");
-    g_feedlistnum = tclmax(rp ? tcatoi(rp) : 10, 1);
-    rp = tctmplconf(g_tmpl, "filenum");
-    g_filenum = tclmax(rp ? tcatoi(rp) : 10, 1);
-    rp = tctmplconf(g_tmpl, "sidebarnum");
-    g_sidebarnum = tclmax(rp ? tcatoi(rp) : 0, 0);
-    g_commentmode = tctmplconf(g_tmpl, "commentmode");
-    if(!g_commentmode) g_commentmode = "";
-    g_updatecmd = tctmplconf(g_tmpl, "updatecmd");
-    if(!g_updatecmd) g_updatecmd = "";
-    rp = tctmplconf(g_tmpl, "sessionlife");
-    g_sessionlife = tclmax(rp ? tcatoi(rp) : 0, 0);
-    g_frontpage = tctmplconf(g_tmpl, "frontpage");
-    if(!g_frontpage) g_frontpage = "";
-    TCMAP *conf = g_tmpl->conf;
-    tcmapiterinit(conf);
-    while((rp = tcmapiternext2(conf)) != NULL){
-      const char *pv = tcmapiterval2(rp);
-      char *name = tcsprintf("TP_%s", rp);
-      tcstrtoupper(name);
-      setenv(name, pv, 1);
-      tcfree(name);
-    }
-    if(g_scrextproc){
-      scrextsetmapvar(g_scrextproc, "_conf", conf);
-      scrextload(g_scrextproc, g_scrext);
-    }
-    eventloop();
-    if(g_scrextproc) scrextdel(g_scrextproc);
-    tcfree(uploadpubbuf);
-    if(g_users) tcmapdel(g_users);
-  } else {
-    showerror(500, "The template file is missing.");
   }
-  tcfree(scriptnamebuf);
-  tcfree(docrootbuf);
-  tctmpldel(g_tmpl);
-  tcfree(tmplpath);
-  tcfree(barepath);
-  tcfree(prefix);
+  tcmpooldel(mpool);
   return 0;
 }
 
@@ -254,14 +268,6 @@ static bool writepasswd(void){
 }
 
 
-/* event loop */
-static void eventloop(void){
-  TCMPOOL *mpool = tcmpoolnew();
-  dosession(mpool);
-  tcmpooldel(mpool);
-}
-
-
 /* process each session */
 static void dosession(TCMPOOL *mpool){
   // download a file
@@ -1334,6 +1340,7 @@ static void dosession(TCMPOOL *mpool){
     if(g_sessionlife > 0) printf(" max-age=%d;", g_sessionlife);
     printf("\r\n");
   }
+  if(g_eventcount > 0) printf("X-Event-Count: %lu\r\n", g_eventcount);
   printf("\r\n");
   fwrite(tmplstr, 1, strlen(tmplstr), stdout);
   fflush(stdout);
